"""
This module contains every scope used in the application.

The format for defining a scope is:
    <SCOPE_SECTION>_<SCOPE_NAME> = "<scope_section>:<scope_name>"
    CLIENT_CREATE = "client:create"

`SCOPE_REGISTRY` is intended as a comprehensive list of all available scopes.
"""

from enum import Enum

AUTHORIZE = "authorize"
CLI_OBJECTS = "cli-objects"
CLIENT = "client"
CONFIG = "config"
CONNECTION = "connection"
CONNECTION_TYPE = "connection_type"
CONNECTOR_TEMPLATE = "connector_template"
CONSENT = "consent"
CONSENT_SETTINGS = "consent_settings"
CREATE = "create"
CREATE_OR_UPDATE = "create_or_update"
CTL_DATASET = "ctl_dataset"
CTL_POLICY = "ctl_policy"
CURRENT_PRIVACY_PREFERENCE = "current-privacy-preference"
DATABASE = "database"
DATA_CATEGORY = "data_category"
DATA_SUBJECT = "data_subject"
DATA_USE = "data_use"
DATASET = "dataset"
DELETE = "delete"
ENCRYPTION = "encryption"
MESSAGING_TEMPLATE = "messaging-template"
EVALUATION = "evaluation"
EXEC = "exec"
FIDES_TAXONOMY = "fides_taxonomy"
GENERATE = "generate"
INSTANTIATE = "instantiate"
MASKING = "masking"
MESSAGING = "messaging"
ORGANIZATION = "organization"
PASSWORD_RESET = "password-reset"
POLICY = "policy"
PRIVACY_EXPERIENCE = "privacy-experience"
PRIVACY_NOTICE = "privacy-notice"
PRIVACY_PREFERENCE_HISTORY = "privacy-preference-history"
PRIVACY_REQUEST = "privacy-request"
PRIVACY_REQUEST_ACCESS_RESULTS = "privacy-request-access-results"
PRIVACY_REQUEST_NOTIFICATIONS = "privacy-request-notifications"
READ = "read"
REGISTER = "register"
RESET = "reset"
RESUME = "resume"
REVIEW = "review"
RULE = "rule"
SAAS_CONFIG = "saas_config"
SCOPE = "scope"
STORAGE = "storage"
SYSTEM = "system"
SYSTEM_MANAGER = "system_manager"
TAXONOMY = "taxonomy"
TEST = "test"
TRANSFER = "transfer"
UPDATE = "update"
UPLOAD_DATA = "upload_data"
USER = "user"
USER_PERMISSION = "user-permission"
VALIDATE = "validate"
VIEW_DATA = "view_data"
WEBHOOK = "webhook"
WORKER_STATS = "worker-stats"

ASSIGN_OWNERS = "assign_owners"

CLIENT_CREATE = f"{CLIENT}:{CREATE}"
CLIENT_DELETE = f"{CLIENT}:{DELETE}"
CLIENT_READ = f"{CLIENT}:{READ}"
CLIENT_UPDATE = f"{CLIENT}:{UPDATE}"

CLI_OBJECTS_CREATE = f"{CLI_OBJECTS}:{CREATE}"
CLI_OBJECTS_READ = f"{CLI_OBJECTS}:{READ}"
CLI_OBJECTS_UPDATE = f"{CLI_OBJECTS}:{UPDATE}"
CLI_OBJECTS_DELETE = f"{CLI_OBJECTS}:{DELETE}"

CONFIG_READ = f"{CONFIG}:{READ}"
CONFIG_UPDATE = f"{CONFIG}:{UPDATE}"

CONNECTION_CREATE_OR_UPDATE = f"{CONNECTION}:{CREATE_OR_UPDATE}"
CONNECTION_DELETE = f"{CONNECTION}:{DELETE}"
CONNECTION_READ = f"{CONNECTION}:{READ}"
CONNECTION_AUTHORIZE = f"{CONNECTION}:{AUTHORIZE}"

CONNECTION_TYPE_READ = f"{CONNECTION_TYPE}:{READ}"

CONSENT_READ = f"{CONSENT}:{READ}"

CONSENT_SETTINGS_READ = f"{CONSENT_SETTINGS}:{READ}"
CONSENT_SETTINGS_UPDATE = f"{CONSENT_SETTINGS}:{UPDATE}"

CTL_DATASET_CREATE = f"{CTL_DATASET}:{CREATE}"
CTL_DATASET_READ = f"{CTL_DATASET}:{READ}"
CTL_DATASET_UPDATE = f"{CTL_DATASET}:{UPDATE}"
CTL_DATASET_DELETE = f"{CTL_DATASET}:{DELETE}"

CTL_POLICY_CREATE = f"{CTL_POLICY}:{CREATE}"
CTL_POLICY_READ = f"{CTL_POLICY}:{READ}"
CTL_POLICY_UPDATE = f"{CTL_POLICY}:{UPDATE}"
CTL_POLICY_DELETE = f"{CTL_POLICY}:{DELETE}"

CURRENT_PRIVACY_PREFERENCE_READ = f"{CURRENT_PRIVACY_PREFERENCE}:{READ}"

DATABASE_RESET = f"{DATABASE}:{RESET}"

DATA_CATEGORY_CREATE = f"{DATA_CATEGORY}:{CREATE}"
DATA_CATEGORY_READ = f"{DATA_CATEGORY}:{READ}"
DATA_CATEGORY_UPDATE = f"{DATA_CATEGORY}:{UPDATE}"
DATA_CATEGORY_DELETE = f"{DATA_CATEGORY}:{DELETE}"


DATA_SUBJECT_CREATE = f"{DATA_SUBJECT}:{CREATE}"
DATA_SUBJECT_READ = f"{DATA_SUBJECT}:{READ}"
DATA_SUBJECT_UPDATE = f"{DATA_SUBJECT}:{UPDATE}"
DATA_SUBJECT_DELETE = f"{DATA_SUBJECT}:{DELETE}"

DATA_USE_CREATE = f"{DATA_USE}:{CREATE}"
DATA_USE_READ = f"{DATA_USE}:{READ}"
DATA_USE_UPDATE = f"{DATA_USE}:{UPDATE}"
DATA_USE_DELETE = f"{DATA_USE}:{DELETE}"

DATASET_CREATE_OR_UPDATE = f"{DATASET}:{CREATE_OR_UPDATE}"
DATASET_DELETE = f"{DATASET}:{DELETE}"
DATASET_READ = f"{DATASET}:{READ}"
DATASET_TEST = f"{DATASET}:{TEST}"

ENCRYPTION_EXEC = f"{ENCRYPTION}:{EXEC}"

EVALUATION_CREATE = f"{EVALUATION}:{CREATE}"
EVALUATION_READ = f"{EVALUATION}:{READ}"
EVALUATION_UPDATE = f"{EVALUATION}:{UPDATE}"
EVALUATION_DELETE = f"{EVALUATION}:{DELETE}"

FIDES_TAXONOMY_UPDATE = f"{FIDES_TAXONOMY}:{UPDATE}"

GENERATE_EXEC = f"{GENERATE}:{EXEC}"

MASKING_EXEC = f"{MASKING}:{EXEC}"
MASKING_READ = f"{MASKING}:{READ}"

MESSAGING_CREATE_OR_UPDATE = f"{MESSAGING}:{CREATE_OR_UPDATE}"
MESSAGING_DELETE = f"{MESSAGING}:{DELETE}"
MESSAGING_READ = f"{MESSAGING}:{READ}"
MESSAGING_TEMPLATE_UPDATE = f"{MESSAGING_TEMPLATE}:{UPDATE}"

ORGANIZATION_CREATE = f"{ORGANIZATION}:{CREATE}"
ORGANIZATION_READ = f"{ORGANIZATION}:{READ}"
ORGANIZATION_UPDATE = f"{ORGANIZATION}:{UPDATE}"
ORGANIZATION_DELETE = f"{ORGANIZATION}:{DELETE}"

POLICY_CREATE_OR_UPDATE = f"{POLICY}:{CREATE_OR_UPDATE}"
POLICY_DELETE = f"{POLICY}:{DELETE}"
POLICY_READ = f"{POLICY}:{READ}"

PRIVACY_EXPERIENCE_CREATE = f"{PRIVACY_EXPERIENCE}:{CREATE}"
PRIVACY_EXPERIENCE_UPDATE = f"{PRIVACY_EXPERIENCE}:{UPDATE}"
PRIVACY_EXPERIENCE_READ = f"{PRIVACY_EXPERIENCE}:{READ}"

PRIVACY_NOTICE_CREATE = f"{PRIVACY_NOTICE}:{CREATE}"
PRIVACY_NOTICE_UPDATE = f"{PRIVACY_NOTICE}:{UPDATE}"
PRIVACY_NOTICE_READ = f"{PRIVACY_NOTICE}:{READ}"

PRIVACY_PREFERENCE_HISTORY_READ = f"{PRIVACY_PREFERENCE_HISTORY}:{READ}"

PRIVACY_REQUEST_CALLBACK_RESUME = f"{PRIVACY_REQUEST}:{RESUME}"  # User has permission to restart a paused privacy request
PRIVACY_REQUEST_CREATE = f"{PRIVACY_REQUEST}:{CREATE}"
PRIVACY_REQUEST_DELETE = f"{PRIVACY_REQUEST}:{DELETE}"
PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE = (
    f"{PRIVACY_REQUEST_NOTIFICATIONS}:{CREATE_OR_UPDATE}"
)
PRIVACY_REQUEST_NOTIFICATIONS_READ = f"{PRIVACY_REQUEST_NOTIFICATIONS}:{READ}"
PRIVACY_REQUEST_READ = f"{PRIVACY_REQUEST}:{READ}"
PRIVACY_REQUEST_REVIEW = f"{PRIVACY_REQUEST}:{REVIEW}"
PRIVACY_REQUEST_TRANSFER = f"{PRIVACY_REQUEST}:{TRANSFER}"
PRIVACY_REQUEST_UPLOAD_DATA = f"{PRIVACY_REQUEST}:{UPLOAD_DATA}"
PRIVACY_REQUEST_VIEW_DATA = f"{PRIVACY_REQUEST}:{VIEW_DATA}"
PRIVACY_REQUEST_READ_ACCESS_RESULTS = f"{PRIVACY_REQUEST_ACCESS_RESULTS}:{READ}"

RULE_CREATE_OR_UPDATE = f"{RULE}:{CREATE_OR_UPDATE}"
RULE_DELETE = f"{RULE}:{DELETE}"
RULE_READ = f"{RULE}:{READ}"

SAAS_CONFIG_CREATE_OR_UPDATE = f"{SAAS_CONFIG}:{CREATE_OR_UPDATE}"
SAAS_CONFIG_DELETE = f"{SAAS_CONFIG}:{DELETE}"
SAAS_CONFIG_READ = f"{SAAS_CONFIG}:{READ}"

SAAS_CONNECTION_INSTANTIATE = f"{CONNECTION}:{INSTANTIATE}"
CONNECTOR_TEMPLATE_REGISTER = f"{CONNECTOR_TEMPLATE}:{REGISTER}"

SCOPE_READ = f"{SCOPE}:{READ}"

STORAGE_CREATE_OR_UPDATE = f"{STORAGE}:{CREATE_OR_UPDATE}"
STORAGE_DELETE = f"{STORAGE}:{DELETE}"
STORAGE_READ = f"{STORAGE}:{READ}"

SYSTEM_CREATE = f"{SYSTEM}:{CREATE}"
SYSTEM_READ = f"{SYSTEM}:{READ}"
SYSTEM_UPDATE = f"{SYSTEM}:{UPDATE}"
SYSTEM_DELETE = f"{SYSTEM}:{DELETE}"


SYSTEM_MANAGER_READ = f"{SYSTEM_MANAGER}:{READ}"
SYSTEM_MANAGER_UPDATE = f"{SYSTEM_MANAGER}:{UPDATE}"
SYSTEM_MANAGER_DELETE = f"{SYSTEM_MANAGER}:{DELETE}"

TAXONOMY_CREATE = f"{TAXONOMY}:{CREATE}"
TAXONOMY_UPDATE = f"{TAXONOMY}:{UPDATE}"
TAXONOMY_DELETE = f"{TAXONOMY}:{DELETE}"

USER_CREATE = f"{USER}:{CREATE}"
USER_DELETE = f"{USER}:{DELETE}"
USER_READ = f"{USER}:{READ}"
USER_UPDATE = f"{USER}:{UPDATE}"
USER_PASSWORD_RESET = f"{USER}:{PASSWORD_RESET}"

USER_PERMISSION_CREATE = f"{USER_PERMISSION}:{CREATE}"
USER_PERMISSION_UPDATE = f"{USER_PERMISSION}:{UPDATE}"
USER_PERMISSION_READ = f"{USER_PERMISSION}:{READ}"
USER_PERMISSION_ASSIGN_OWNERS = f"{USER_PERMISSION}:{ASSIGN_OWNERS}"

VALIDATE_EXEC = f"{VALIDATE}:{EXEC}"

WEBHOOK_CREATE_OR_UPDATE = f"{WEBHOOK}:{CREATE_OR_UPDATE}"
WEBHOOK_DELETE = f"{WEBHOOK}:{DELETE}"
WEBHOOK_READ = f"{WEBHOOK}:{READ}"

WORKER_STATS_READ = f"{WORKER_STATS}:{READ}"

SCOPE_DOCS = {
    CONFIG_READ: "View the configuration",
    CONFIG_UPDATE: "Update the configuration",
    CLI_OBJECTS_CREATE: "Create objects via the command line interface",
    CLI_OBJECTS_READ: "Read objects via the command line interface",
    CLI_OBJECTS_UPDATE: "Update objects via the command line interface",
    CLI_OBJECTS_DELETE: "Delete objects via the command line interface",
    CLIENT_CREATE: "Create OAuth clients",
    CLIENT_DELETE: "Remove OAuth clients",
    CLIENT_READ: "View current scopes for OAuth clients",
    CLIENT_UPDATE: "Modify existing scopes for OAuth clients",
    CONNECTION_CREATE_OR_UPDATE: "Create or modify connections",
    CONNECTION_DELETE: "Remove connections",
    CONNECTION_READ: "View connections",
    CONNECTION_AUTHORIZE: "OAuth2 Authorization",
    CONNECTION_TYPE_READ: "View types of connections",
    CONNECTOR_TEMPLATE_REGISTER: "Register a connector template",
    CONSENT_READ: "Read consent preferences",
    CONSENT_SETTINGS_READ: "Read org-wide consent settings",
    CONSENT_SETTINGS_UPDATE: "Update org-wide consent settings",
    CTL_DATASET_CREATE: "Create a ctl dataset",
    CTL_DATASET_READ: "Read ctl datasets",
    CTL_DATASET_DELETE: "Delete a ctl dataset",
    CTL_DATASET_UPDATE: "Update ctl datasets",
    CTL_POLICY_CREATE: "Create a ctl policy",
    CTL_POLICY_READ: "Read ctl policies",
    CTL_POLICY_DELETE: "Delete a ctl policy",
    CTL_POLICY_UPDATE: "Update ctl policies",
    CURRENT_PRIVACY_PREFERENCE_READ: "Read the current privacy preferences of all users",
    DATABASE_RESET: "Reset the application database",
    DATA_CATEGORY_CREATE: "Create a data category",
    DATA_CATEGORY_DELETE: "Delete data categories",
    DATA_CATEGORY_READ: "Read data categories",
    DATA_CATEGORY_UPDATE: "Update data categories",
    DATA_SUBJECT_CREATE: "Create a data subject",
    DATA_SUBJECT_READ: "Read data subjects",
    DATA_SUBJECT_DELETE: "Delete data subjects",
    DATA_SUBJECT_UPDATE: "Update data subjects",
    DATA_USE_CREATE: "Create a data use",
    DATA_USE_READ: "Read data uses",
    DATA_USE_DELETE: "Delete data uses",
    DATA_USE_UPDATE: "Update data uses",
    DATASET_CREATE_OR_UPDATE: "Create or modify datasets",
    DATASET_DELETE: "Delete datasets",
    DATASET_READ: "View datasets",
    DATASET_TEST: "Run a standalone privacy request test for a dataset",
    ENCRYPTION_EXEC: "Encrypt data",
    MESSAGING_TEMPLATE_UPDATE: "Update messaging templates",
    EVALUATION_CREATE: "Create evaluation",
    EVALUATION_READ: "Read evaluations",
    EVALUATION_DELETE: "Delete evaluations",
    EVALUATION_UPDATE: "Update evaluations",
    FIDES_TAXONOMY_UPDATE: "Update default fides taxonomy description",
    GENERATE_EXEC: "",
    MASKING_EXEC: "Execute a masking strategy",
    MASKING_READ: "Read masking strategies",
    MESSAGING_CREATE_OR_UPDATE: "",
    MESSAGING_DELETE: "",
    MESSAGING_READ: "",
    ORGANIZATION_CREATE: "Create organization",
    ORGANIZATION_READ: "Read organization details",
    ORGANIZATION_DELETE: "Delete organization",
    ORGANIZATION_UPDATE: "Update organization details",
    POLICY_CREATE_OR_UPDATE: "Create or modify policies",
    POLICY_DELETE: "Remove policies",
    POLICY_READ: "View policies",
    PRIVACY_EXPERIENCE_CREATE: "Create privacy experiences",
    PRIVACY_EXPERIENCE_UPDATE: "Update privacy experiences",
    PRIVACY_EXPERIENCE_READ: "View privacy experiences",
    PRIVACY_NOTICE_CREATE: "Create privacy notices",
    PRIVACY_NOTICE_UPDATE: "Update privacy notices",
    PRIVACY_NOTICE_READ: "View privacy notices",
    PRIVACY_PREFERENCE_HISTORY_READ: "Read the history of all saved privacy preferences",
    PRIVACY_REQUEST_CREATE: "",
    PRIVACY_REQUEST_CALLBACK_RESUME: "Restart paused privacy requests",
    PRIVACY_REQUEST_READ_ACCESS_RESULTS: "Download access data for the privacy request",
    PRIVACY_REQUEST_DELETE: "Remove privacy requests",
    PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE: "",
    PRIVACY_REQUEST_NOTIFICATIONS_READ: "",
    PRIVACY_REQUEST_READ: "View privacy requests",
    PRIVACY_REQUEST_REVIEW: "Review privacy requests",
    PRIVACY_REQUEST_TRANSFER: "Transfer privacy requests",
    PRIVACY_REQUEST_UPLOAD_DATA: "Manually upload data for the privacy request",
    PRIVACY_REQUEST_VIEW_DATA: "View subject data related to the privacy request",
    RULE_CREATE_OR_UPDATE: "Create or update rules",
    RULE_DELETE: "Remove rules",
    RULE_READ: "View rules",
    SAAS_CONFIG_CREATE_OR_UPDATE: "Create or update SaaS configurations",
    SAAS_CONFIG_DELETE: "Remove SaaS configurations",
    SAAS_CONFIG_READ: "View SaaS configurations",
    SAAS_CONNECTION_INSTANTIATE: "Instantiate a SaaS connection config from a connector template",
    SCOPE_READ: "View authorization scopes",
    STORAGE_CREATE_OR_UPDATE: "Create or update storage",
    STORAGE_DELETE: "Remove storage",
    STORAGE_READ: "View storage",
    SYSTEM_CREATE: "Create systems",
    SYSTEM_READ: "Read systems",
    SYSTEM_DELETE: "Delete systems",
    SYSTEM_UPDATE: "Update systems",
    SYSTEM_MANAGER_READ: "Read systems users can manage",
    SYSTEM_MANAGER_DELETE: "Delete systems user can manage",
    SYSTEM_MANAGER_UPDATE: "Update systems user can manage",
    TAXONOMY_CREATE: "Create local taxonomy",
    TAXONOMY_DELETE: "Delete local taxonomy",
    TAXONOMY_UPDATE: "Update local taxonomy",
    USER_CREATE: "Create users",
    USER_UPDATE: "Update users",
    USER_DELETE: "Remove users",
    USER_READ: "View users",
    USER_PASSWORD_RESET: "Reset another user's password",
    USER_PERMISSION_CREATE: "Create user permissions",
    USER_PERMISSION_UPDATE: "Update user permissions",
    USER_PERMISSION_ASSIGN_OWNERS: "Assign the owner role to a user",
    USER_PERMISSION_READ: "View user permissions",
    VALIDATE_EXEC: "",
    WEBHOOK_CREATE_OR_UPDATE: "Create or update web hooks",
    WEBHOOK_DELETE: "Remove web hooks",
    WEBHOOK_READ: "View web hooks",
    WORKER_STATS_READ: "View worker statistics",
}

SCOPE_REGISTRY = list(SCOPE_DOCS.keys())

# mypy doesn't like taking the dictionary to generate the enum
# https://github.com/python/mypy/issues/5317
ScopeRegistryEnum = Enum(  # type: ignore[misc]
    "ScopeRegistryEnum", {scope: scope for scope in SCOPE_REGISTRY}  # type: ignore
)



from pydantic import Field
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__CELERY__"


class CelerySettings(FidesSettings):
    """Configuration settings for Celery.  Only a small subset can be configured through Fides env vars"""

    event_queue_prefix: str = Field(
        default="fides_worker",
        description="The prefix to use for event receiver queue names",
    )
    task_default_queue: str = Field(
        default="fides",
        description="The name of the default queue if a message has no route or no custom queue has been specified",
    )
    task_always_eager: bool = Field(
        default=True,
        description="If true, tasks are executed locally instead of being sent to the queue.  "
        "If False, tasks are sent to the queue.",
    )
    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



from typing import Any, Dict

from sqlalchemy.orm import Session

from fides.api.db.base_class import Base


def update_if_modified(resource: Base, db: Session, *, data: Dict[str, Any]) -> bool:
    """Updates a resource and returns whether it was modified.

    This is used when deciding whether to save a new version for a privacy notice or privacy experience config.
    """
    # run through potential updates
    for key, value in data.items():
        setattr(resource, key, value)

    if db.is_modified(resource):
        resource.save(db)
        return True

    return False


def create_historical_data_from_record(resource: Base) -> Dict:
    """Prep data to be saved in a historical table for record keeping

    This is used when creating a new version for a privacy notice or privacy experience config.
    """
    # Resource attributes are being lazy loaded and not showing up in the dict.  Accessing
    # an attribute causes them to be loaded.
    resource.id  # pylint: disable=pointless-statement
    return clone_object_attributes(resource)


def clone_object_attributes(resource: Base) -> Dict:
    """
    Copies an objects attributes into a dictionary.

    Removes protected fields like the SQLAlchemy instance state and id.
    """
    cloned_attributes = resource.__dict__.copy()
    # remove protected fields from the cloned dict
    cloned_attributes.pop("_sa_instance_state", None)
    cloned_attributes.pop("id", None)
    # remove datetime fields
    cloned_attributes.pop("created_at", None)
    cloned_attributes.pop("updated_at", None)
    return cloned_attributes


def dry_update_data(resource: Base, data_updates: Dict) -> Dict:
    """Copy the record as a dictionary and apply patch updates"""
    # Clone original record
    cloned_attributes: Dict = clone_object_attributes(resource)
    # Replace existing fields with supplied updates
    for key, val in data_updates.items():
        cloned_attributes[key] = val
    return cloned_attributes



"""data protection impact assessments additional column on systems

Revision ID: a2b9b156ebaf
Revises: adad3be6af08
Create Date: 2022-04-01 18:42:58.961132

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a2b9b156ebaf"
down_revision = "adad3be6af08"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "systems",
        sa.Column("data_protection_impact_assessment", sa.JSON(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("systems", "data_protection_impact_assessment")
    # ### end Alembic commands ###



"""deprecate_current_pref_v1

Revision ID: 23e01f8f4652
Revises: 9e83545ed9b6
Create Date: 2024-04-29 23:16:41.909736

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "23e01f8f4652"
down_revision = "9e83545ed9b6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_lastservednotice_created_at", table_name="lastservednotice")
    op.drop_index("ix_lastservednotice_feature", table_name="lastservednotice")
    op.drop_index(
        "ix_lastservednotice_fides_user_device_provided_identity_id",
        table_name="lastservednotice",
    )
    op.drop_index("ix_lastservednotice_id", table_name="lastservednotice")
    op.drop_index(
        "ix_lastservednotice_privacy_notice_history_id", table_name="lastservednotice"
    )
    op.drop_index(
        "ix_lastservednotice_privacy_notice_id", table_name="lastservednotice"
    )
    op.drop_index(
        "ix_lastservednotice_provided_identity_id", table_name="lastservednotice"
    )
    op.drop_index("ix_lastservednotice_purpose_consent", table_name="lastservednotice")
    op.drop_index(
        "ix_lastservednotice_purpose_legitimate_interests",
        table_name="lastservednotice",
    )
    op.drop_index("ix_lastservednotice_special_feature", table_name="lastservednotice")
    op.drop_index("ix_lastservednotice_special_purpose", table_name="lastservednotice")
    op.drop_index("ix_lastservednotice_system_consent", table_name="lastservednotice")
    op.drop_index(
        "ix_lastservednotice_system_legitimate_interests", table_name="lastservednotice"
    )
    op.drop_index("ix_lastservednotice_updated_at", table_name="lastservednotice")
    op.drop_index("ix_lastservednotice_vendor_consent", table_name="lastservednotice")
    op.drop_index(
        "ix_lastservednotice_vendor_legitimate_interests", table_name="lastservednotice"
    )
    op.drop_table("lastservednotice")
    op.drop_index(
        "ix_currentprivacypreference_created_at", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_feature", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_fides_user_device_provided__d6ea",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_id", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_preference", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_privacy_notice_history_id",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_privacy_notice_id",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_provided_identity_id",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_purpose_consent",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_purpose_legitimate_interests",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_special_feature",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_special_purpose",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_system_consent",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_system_legitimate_interests",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_updated_at", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_vendor_consent",
        table_name="currentprivacypreference",
    )
    op.drop_index(
        "ix_currentprivacypreference_vendor_legitimate_interests",
        table_name="currentprivacypreference",
    )
    op.drop_table("currentprivacypreference")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "currentprivacypreference",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=False),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("preference", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "provided_identity_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "privacy_notice_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "privacy_notice_history_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "fides_user_device_provided_identity_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("tcf_version", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column("feature", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("special_feature", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("special_purpose", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("purpose_consent", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column(
            "purpose_legitimate_interests",
            sa.INTEGER(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("vendor_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "vendor_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("system_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "system_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["fides_user_device_provided_identity_id"],
            ["providedidentity.id"],
            name="currentprivacypreference_fides_user_device_provided_identi_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_history_id"],
            ["privacynoticehistory.id"],
            name="currentprivacypreference_privacy_notice_history_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_id"],
            ["privacynotice.id"],
            name="currentprivacypreference_privacy_notice_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
            name="currentprivacypreference_provided_identity_id_fkey",
        ),
        sa.PrimaryKeyConstraint("id", name="currentprivacypreference_pkey"),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "feature",
            name="fides_user_device_identity_feature",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "privacy_notice_id",
            name="fides_user_device_identity_privacy_notice",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "purpose_consent",
            name="fides_user_device_identity_purpose_consent",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "purpose_legitimate_interests",
            name="fides_user_device_identity_purpose_leg_interests",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "special_feature",
            name="fides_user_device_identity_special_feature",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "special_purpose",
            name="fides_user_device_identity_special_purpose",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "system_consent",
            name="fides_user_device_identity_system_consent",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "system_legitimate_interests",
            name="fides_user_device_identity_system_leg_interests",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "vendor_consent",
            name="fides_user_device_identity_vendor_consent",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "vendor_legitimate_interests",
            name="fides_user_device_identity_vendor_leg_interests",
        ),
        sa.UniqueConstraint("provided_identity_id", "feature", name="identity_feature"),
        sa.UniqueConstraint(
            "provided_identity_id", "privacy_notice_id", name="identity_privacy_notice"
        ),
        sa.UniqueConstraint(
            "provided_identity_id", "purpose_consent", name="identity_purpose_consent"
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "purpose_legitimate_interests",
            name="identity_purpose_leg_interests",
        ),
        sa.UniqueConstraint(
            "provided_identity_id", "special_feature", name="identity_special_feature"
        ),
        sa.UniqueConstraint(
            "provided_identity_id", "special_purpose", name="identity_special_purpose"
        ),
        sa.UniqueConstraint(
            "provided_identity_id", "system_consent", name="identity_system_consent"
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "system_legitimate_interests",
            name="identity_system_leg_interests",
        ),
        sa.UniqueConstraint(
            "provided_identity_id", "vendor_consent", name="identity_vendor_consent"
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "vendor_legitimate_interests",
            name="identity_vendor_leg_interests",
        ),
    )
    op.create_index(
        "ix_currentprivacypreference_vendor_legitimate_interests",
        "currentprivacypreference",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_vendor_consent",
        "currentprivacypreference",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_updated_at",
        "currentprivacypreference",
        ["updated_at"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_system_legitimate_interests",
        "currentprivacypreference",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_system_consent",
        "currentprivacypreference",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_special_purpose",
        "currentprivacypreference",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_special_feature",
        "currentprivacypreference",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_purpose_legitimate_interests",
        "currentprivacypreference",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_purpose_consent",
        "currentprivacypreference",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_provided_identity_id",
        "currentprivacypreference",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_privacy_notice_id",
        "currentprivacypreference",
        ["privacy_notice_id"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_privacy_notice_history_id",
        "currentprivacypreference",
        ["privacy_notice_history_id"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_preference",
        "currentprivacypreference",
        ["preference"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_id",
        "currentprivacypreference",
        ["id"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_fides_user_device_provided__d6ea",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_feature",
        "currentprivacypreference",
        ["feature"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_created_at",
        "currentprivacypreference",
        ["created_at"],
        unique=False,
    )
    op.create_table(
        "lastservednotice",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=False),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "provided_identity_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "fides_user_device_provided_identity_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "privacy_notice_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "privacy_notice_history_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("tcf_version", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column("feature", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("special_feature", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("special_purpose", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("purpose_consent", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column(
            "purpose_legitimate_interests",
            sa.INTEGER(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("vendor_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "vendor_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("system_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "system_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["fides_user_device_provided_identity_id"],
            ["providedidentity.id"],
            name="lastservednotice_fides_user_device_provided_identity_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_history_id"],
            ["privacynoticehistory.id"],
            name="lastservednotice_privacy_notice_history_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_id"],
            ["privacynotice.id"],
            name="lastservednotice_privacy_notice_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
            name="lastservednotice_provided_identity_id_fkey",
        ),
        sa.PrimaryKeyConstraint("id", name="lastservednotice_pkey"),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "feature",
            name="last_served_fides_user_device_identity_feature",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "privacy_notice_id",
            name="last_served_fides_user_device_identity_privacy_notice",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "purpose_consent",
            name="last_served_fides_user_device_identity_purpose_consent",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "purpose_legitimate_interests",
            name="last_served_fides_user_device_identity_purpose_leg_interests",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "special_feature",
            name="last_served_fides_user_device_identity_special_feature",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "special_purpose",
            name="last_served_fides_user_device_identity_special_purpose",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "system_consent",
            name="last_served_fides_user_device_identity_system_consent",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "system_legitimate_interests",
            name="last_served_fides_user_device_identity_system_leg_interests",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "vendor_consent",
            name="last_served_fides_user_device_identity_vendor_consent",
        ),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "vendor_legitimate_interests",
            name="last_served_fides_user_device_identity_vendor_leg_interests",
        ),
        sa.UniqueConstraint(
            "provided_identity_id", "feature", name="last_served_identity_feature"
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "privacy_notice_id",
            name="last_served_identity_privacy_notice",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "purpose_consent",
            name="last_served_identity_purpose_consent",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "purpose_legitimate_interests",
            name="last_served_identity_purpose_legitimate_interests",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "special_feature",
            name="last_served_identity_special_feature",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "special_purpose",
            name="last_served_identity_special_purpose",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "system_consent",
            name="last_served_identity_system_consent",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "system_legitimate_interests",
            name="last_served_identity_system_leg_interests",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "vendor_consent",
            name="last_served_identity_vendor_consent",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "vendor_legitimate_interests",
            name="last_served_identity_vendor_leg_interests",
        ),
    )
    op.create_index(
        "ix_lastservednotice_vendor_legitimate_interests",
        "lastservednotice",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_vendor_consent",
        "lastservednotice",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_updated_at",
        "lastservednotice",
        ["updated_at"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_system_legitimate_interests",
        "lastservednotice",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_system_consent",
        "lastservednotice",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_special_purpose",
        "lastservednotice",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_special_feature",
        "lastservednotice",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_purpose_legitimate_interests",
        "lastservednotice",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_purpose_consent",
        "lastservednotice",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_provided_identity_id",
        "lastservednotice",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_privacy_notice_id",
        "lastservednotice",
        ["privacy_notice_id"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_privacy_notice_history_id",
        "lastservednotice",
        ["privacy_notice_history_id"],
        unique=False,
    )
    op.create_index("ix_lastservednotice_id", "lastservednotice", ["id"], unique=False)
    op.create_index(
        "ix_lastservednotice_fides_user_device_provided_identity_id",
        "lastservednotice",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_lastservednotice_feature", "lastservednotice", ["feature"], unique=False
    )
    op.create_index(
        "ix_lastservednotice_created_at",
        "lastservednotice",
        ["created_at"],
        unique=False,
    )
    # ### end Alembic commands ###



from typing import Any, Dict

from fides.api.schemas.connection_configuration.connection_secrets_sovrn import (
    SOVRN_REQUIRED_IDENTITY,
)
from fides.api.service.connectors.consent_email_connector import (
    GenericConsentEmailConnector,
)


class SovrnConnector(GenericConsentEmailConnector):
    """SovrnConnector - only need to override the details for the test email."""

    @property
    def identities_for_test_email(self) -> Dict[str, Any]:
        return {SOVRN_REQUIRED_IDENTITY: "test_ljt_reader_id"}



from __future__ import annotations

import enum
from datetime import datetime
from typing import TYPE_CHECKING, Any, List, Optional, Type

from loguru import logger
from sqlalchemy import Boolean, Column, DateTime, Enum, ForeignKey, String, event
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import RelationshipProperty, Session, relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.common_exceptions import KeyOrNameAlreadyExists
from fides.api.db.base_class import Base, FidesBase, JSONTypeOverride
from fides.api.models.consent_automation import ConsentAutomation
from fides.api.models.sql_models import System  # type: ignore[attr-defined]
from fides.api.schemas.policy import ActionType
from fides.api.schemas.saas.saas_config import SaaSConfig
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.detection_discovery import MonitorConfig
    from fides.api.schemas.connection_configuration.enums.system_type import SystemType


class ConnectionTestStatus(enum.Enum):
    """Enum for supplying statuses of validating credentials for a Connection Config to the user"""

    succeeded = "succeeded"
    failed = "failed"
    skipped = "skipped"


class ConnectionType(enum.Enum):
    """
    Supported types to which we can connect Fides.
    """

    attentive_email = "attentive_email"
    bigquery = "bigquery"
    datahub = "datahub"
    dynamodb = "dynamodb"
    fides = "fides"
    generic_consent_email = "generic_consent_email"  # Run after the traversal
    generic_erasure_email = "generic_erasure_email"  # Run after the traversal
    dynamic_erasure_email = "dynamic_erasure_email"  # Run after the traversal
    google_cloud_sql_mysql = "google_cloud_sql_mysql"
    google_cloud_sql_postgres = "google_cloud_sql_postgres"
    https = "https"
    manual = "manual"  # Deprecated - use manual_webhook instead
    manual_webhook = "manual_webhook"  # Runs upfront before the traversal
    mariadb = "mariadb"
    mongodb = "mongodb"
    mssql = "mssql"
    mysql = "mysql"
    postgres = "postgres"
    rds_mysql = "rds_mysql"
    rds_postgres = "rds_postgres"
    redshift = "redshift"
    s3 = "s3"
    saas = "saas"
    scylla = "scylla"
    snowflake = "snowflake"
    sovrn = "sovrn"
    timescale = "timescale"
    website = "website"

    @property
    def human_readable(self) -> str:
        """Human-readable mapping for ConnectionTypes
        Add to this mapping if you add a new ConnectionType
        """
        readable_mapping: dict[str, str] = {
            ConnectionType.attentive_email.value: "Attentive Email",
            ConnectionType.bigquery.value: "BigQuery",
            ConnectionType.datahub.value: "DataHub",
            ConnectionType.dynamic_erasure_email.value: "Dynamic Erasure Email",
            ConnectionType.dynamodb.value: "DynamoDB",
            ConnectionType.fides.value: "Fides Connector",
            ConnectionType.generic_consent_email.value: "Generic Consent Email",
            ConnectionType.generic_erasure_email.value: "Generic Erasure Email",
            ConnectionType.google_cloud_sql_mysql.value: "Google Cloud SQL for MySQL",
            ConnectionType.google_cloud_sql_postgres.value: "Google Cloud SQL for Postgres",
            ConnectionType.https.value: "Policy Webhook",
            ConnectionType.manual_webhook.value: "Manual Process",
            ConnectionType.manual.value: "Manual Connector",
            ConnectionType.mariadb.value: "MariaDB",
            ConnectionType.mongodb.value: "MongoDB",
            ConnectionType.mssql.value: "Microsoft SQL Server",
            ConnectionType.mysql.value: "MySQL",
            ConnectionType.postgres.value: "PostgreSQL",
            ConnectionType.rds_mysql.value: "RDS MySQL",
            ConnectionType.rds_postgres.value: "RDS Postgres",
            ConnectionType.redshift.value: "Amazon Redshift",
            ConnectionType.s3.value: "Amazon S3",
            ConnectionType.saas.value: "SaaS",
            ConnectionType.scylla.value: "Scylla DB",
            ConnectionType.snowflake.value: "Snowflake",
            ConnectionType.sovrn.value: "Sovrn",
            ConnectionType.timescale.value: "TimescaleDB",
            ConnectionType.website.value: "Website",
        }
        try:
            return readable_mapping[self.value]
        except KeyError:
            raise NotImplementedError(
                "Add new ConnectionType to human_readable mapping"
            )

    @property
    def system_type(self) -> "SystemType":
        from fides.api.schemas.connection_configuration.enums.system_type import (
            SystemType,
        )

        system_type_mapping: dict[str, SystemType] = {
            ConnectionType.attentive_email.value: SystemType.email,
            ConnectionType.bigquery.value: SystemType.database,
            ConnectionType.datahub.value: SystemType.data_catalog,
            ConnectionType.dynamic_erasure_email.value: SystemType.email,
            ConnectionType.dynamodb.value: SystemType.database,
            ConnectionType.fides.value: SystemType.manual,
            ConnectionType.generic_consent_email.value: SystemType.email,
            ConnectionType.generic_erasure_email.value: SystemType.email,
            ConnectionType.google_cloud_sql_mysql.value: SystemType.database,
            ConnectionType.google_cloud_sql_postgres.value: SystemType.database,
            ConnectionType.https.value: SystemType.manual,
            ConnectionType.manual_webhook.value: SystemType.manual,
            ConnectionType.manual.value: SystemType.manual,
            ConnectionType.mariadb.value: SystemType.database,
            ConnectionType.mongodb.value: SystemType.database,
            ConnectionType.mssql.value: SystemType.database,
            ConnectionType.mysql.value: SystemType.database,
            ConnectionType.postgres.value: SystemType.database,
            ConnectionType.rds_mysql.value: SystemType.database,
            ConnectionType.rds_postgres.value: SystemType.database,
            ConnectionType.redshift.value: SystemType.database,
            ConnectionType.s3.value: SystemType.database,
            ConnectionType.saas.value: SystemType.saas,
            ConnectionType.scylla.value: SystemType.database,
            ConnectionType.snowflake.value: SystemType.database,
            ConnectionType.sovrn.value: SystemType.email,
            ConnectionType.timescale.value: SystemType.database,
            ConnectionType.website.value: SystemType.website,
        }

        try:
            return system_type_mapping[self.value]
        except KeyError:
            raise NotImplementedError("Add new ConnectionType to system_type mapping")


class AccessLevel(enum.Enum):
    """
    Perms given to the ConnectionConfig.  For example, with "read" permissions, fidesops promises
    to not modify the data on a connected application database in any way.

    "Write" perms mean we can update/delete items in the connected database.
    """

    read = "read"
    write = "write"


class ConnectionConfig(Base):
    """
    Stores credentials to connect fidesops to an engineer's application databases.
    """

    name = Column(String, nullable=True)
    key = Column(String, index=True, unique=True, nullable=False)
    description = Column(String, index=True, nullable=True)
    connection_type = Column(Enum(ConnectionType), nullable=False)
    access = Column(Enum(AccessLevel), nullable=False)
    secrets = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db
    last_test_timestamp = Column(DateTime(timezone=True))
    last_test_succeeded = Column(Boolean)
    disabled = Column(Boolean, server_default="f", default=False)
    disabled_at = Column(DateTime(timezone=True))

    # Optional column to store the last time the connection was "ran"
    # Each integration can determine the semantics of what "being run" is
    last_run_timestamp = Column(DateTime(timezone=True), nullable=True)

    # only applicable to ConnectionConfigs of connection type saas
    saas_config = Column(
        MutableDict.as_mutable(JSONB), index=False, unique=False, nullable=True
    )

    system_id = Column(
        String, ForeignKey(System.id_field_path), nullable=True, index=True
    )

    datasets = relationship(  # type: ignore[misc]
        "DatasetConfig",
        back_populates="connection_config",
        cascade="all, delete",
    )

    # Monitor configs related to this connection config.
    # If the connection config is deleted, the monitor configs will be deleted as well.
    monitors: RelationshipProperty[List["MonitorConfig"]] = relationship(
        "MonitorConfig",
        back_populates="connection_config",
        cascade="all, delete",
    )

    access_manual_webhook = relationship(  # type: ignore[misc]
        "AccessManualWebhook",
        back_populates="connection_config",
        cascade="delete",
        uselist=False,
    )

    pre_approval_webhooks = relationship(  # type: ignore[misc]
        "PreApprovalWebhook",
        back_populates="connection_config",
        cascade="delete",
    )

    system = relationship(System, back_populates="connection_configs", uselist=False)

    consent_automation: RelationshipProperty[Optional[ConsentAutomation]] = (
        relationship(ConsentAutomation, uselist=False, cascade="all, delete-orphan")
    )

    # Identifies the privacy actions needed from this connection by the associated system.
    enabled_actions = Column(
        ARRAY(Enum(ActionType, native_enum=False)), unique=False, nullable=True
    )

    @property
    def system_key(self) -> Optional[str]:
        """Property for caching a system identifier for systems (or connector names as a fallback) for consent reporting"""
        if self.system:
            return self.system.fides_key
        # TODO: Remove this fallback once all connection configs are linked to systems
        # This will always be None in the future. `self.system` will always be set.
        return self.name

    @property
    def authorized(self) -> bool:
        """Returns True if the connection config has an access token, used for OAuth2 connections"""

        saas_config = self.get_saas_config()
        if not saas_config:
            return False

        authentication = saas_config.client_config.authentication
        if not authentication:
            return False

        # hard-coding to avoid cyclic dependency
        if authentication.strategy not in [
            "oauth2_authorization_code",
            "oauth2_client_credentials",
        ]:
            return False

        return bool(self.secrets and "access_token" in self.secrets.keys())

    @property
    def name_or_key(self) -> str:
        """Returns the ConnectionConfig name if it exists, or its key otherwise."""
        return self.name or self.key

    @classmethod
    def create_without_saving(
        cls: Type[ConnectionConfig], db: Session, *, data: dict[str, Any]
    ) -> ConnectionConfig:
        """Create a ConnectionConfig without persisting to the database"""
        # Build properly formatted key/name for ConnectionConfig.
        # Borrowed from OrmWrappedFidesBase.create
        if hasattr(cls, "key"):
            if db.query(cls).filter_by(key=data["key"]).first():
                raise KeyOrNameAlreadyExists(
                    f"Key {data['key']} already exists in {cls.__name__}. Keys will be snake-cased names if not provided. "
                    f"If you are seeing this error without providing a key, please provide a key or a different name."
                    ""
                )

        # Create
        db_obj = cls(**data)  # type: ignore
        return db_obj

    def get_saas_config(self) -> Optional[SaaSConfig]:
        """Returns a SaaSConfig object from a yaml config"""
        return SaaSConfig(**self.saas_config) if self.saas_config else None

    def update_saas_config(
        self,
        db: Session,
        saas_config: SaaSConfig,
    ) -> None:
        """
        Updates the SaaS config and initializes any empty secrets with
        connector param default values if available (will not override any existing secrets)
        """
        default_secrets = {
            connector_param.name: connector_param.default_value
            for connector_param in saas_config.connector_params
            if connector_param.default_value
        }
        updated_secrets = {**default_secrets, **(self.secrets or {})}
        self.secrets = updated_secrets
        self.saas_config = saas_config.model_dump(mode="json")
        self.save(db)

    def update_test_status(
        self, test_status: ConnectionTestStatus, db: Session
    ) -> None:
        """Updates last_test_timestamp and last_test_succeeded after an attempt to make a test connection.

        If the test was skipped, for example, on an HTTP Connector, don't update these fields.
        """
        if test_status == ConnectionTestStatus.skipped:
            return

        self.last_test_timestamp = datetime.now()
        self.last_test_succeeded = test_status == ConnectionTestStatus.succeeded
        self.save(db)

    def delete(self, db: Session) -> Optional[FidesBase]:
        """Hard deletes datastores that map this ConnectionConfig."""
        logger.info(
            "Deleting connection config {}...",
            self.key,
        )
        for dataset in self.datasets:
            dataset.delete(db=db)

        for monitor in self.monitors:
            logger.info(
                "Deleting monitor config {} associated with connection config {}...",
                monitor.key,
                self.key,
            )
            monitor.delete(db=db)

        return super().delete(db=db)


@event.listens_for(ConnectionConfig.disabled, "set")
def connection_config_disabled_set(
    target: ConnectionConfig, value: bool, original_value: bool, _: Any
) -> None:
    """Update ConnectionConfig.disabled_at if ConnectionConfig.disabled changes"""
    if value != original_value:
        target.disabled_at = datetime.utcnow() if value else None
